import functools
import json
import os
import subprocess
from copy import deepcopy
from typing import Any, Dict, List

from chaoslib.exceptions import InterruptExecution
from chaoslib.types import (
    Activity,
    Configuration,
    Experiment,
    Hypothesis,
    Journal,
    Run,
    Secrets,
    Settings,
)


def singleton(cls):
    """Creates a singleton wrapper for any class"""

    @functools.wraps(cls)
    def wrapper_singleton(*args, **kwargs):
        if not wrapper_singleton.instance_:
            wrapper_singleton.instance_ = cls(*args, **kwargs)
        return wrapper_singleton.instance_

    wrapper_singleton.instance_ = None
    return wrapper_singleton


@singleton
class Terraform:
    def __init__(self):
        super().__init__()
        self.retain = False
        self.silent = False
        self.chdir = None
        self.args = {}

    def configure(
        self,
        retain: bool = False,
        silent: bool = False,
        chdir: str = None,
        args: Dict = None,
    ):
        self.retain = retain
        self.silent = silent
        self.chdir = chdir
        self.args = args or {}

    @property
    def _terraform(self):
        if self.chdir:
            if not os.path.exists(self.chdir):
                raise InterruptExecution(
                    f"Terraform: chdir [{self.chdir}] does not exists"
                )
            if not os.path.isdir(self.chdir):
                raise InterruptExecution(
                    f"Terraform: chdir [{self.chdir}] is not a directory"
                )
            return f"terraform -chdir={self.chdir}"
        return "terraform"

    def terraform_init(self):
        if not os.path.exists(".terraform"):
            result = subprocess.run(
                f"{self._terraform} init",
                capture_output=self.silent,
                shell=True,
            )
            if result.returncode != 0:
                raise InterruptExecution("Failed to initialize terraform")

    def apply(self, **kwargs):
        args = deepcopy(self.args)
        args.update(kwargs)

        var_overrides = []
        for key, value in args.items():
            string_value = value
            if isinstance(value, bool):
                string_value = str(value).lower()
            var_overrides.append(f"-var {key}='{string_value}'")
        opts = " ".join(var_overrides)

        result = subprocess.run(
            f"{self._terraform} apply {opts} -auto-approve",
            capture_output=self.silent,
            shell=True,
        )
        if result.returncode != 0:
            raise InterruptExecution("Failed to apply terraform stack terraform")

    def output(self):
        result = subprocess.run(
            f"{self._terraform} output -json",
            shell=True,
            capture_output=True,
            text=True,
        )
        outputs = json.loads(result.stdout)
        return outputs

    def destroy(self):
        subprocess.run(
            f"{self._terraform} destroy -auto-approve",
            capture_output=self.silent,
            shell=True,
        )



"""Top-level package for chaostf"""
from typing import List

from chaoslib.discovery.discover import (
    discover_actions,
    discover_activities,
    discover_probes,
    initialize_discovery_result,
)
from chaoslib.types import DiscoveredActivities, Discovery
from logzero import logger

name = "chaostf"
__author__ = """Manuel Castellin"""
__email__ = "manuel@castellinconsulting.com"
__version__ = "0.0.4"
__all__ = [
    "discover",
    "__version__",
]


def discover(discover_system: bool = True) -> Discovery:
    # pylint: disable=unused-argument
    discovery = initialize_discovery_result("chaostf", __version__, "chaostf")
    discovery["activities"].extend(load_exported_activities())
    return discovery


def load_exported_activities() -> List[DiscoveredActivities]:
    """
    Extract metadata from actions and probes exposed by this extension.
    """
    activities = []
    activities.extend(discover_actions("chaostf.actions"))
    activities.extend(discover_activities("chaostf.control", "control"))

    return activities



from typing import Any, Dict, List, Union

from chaoslib.exceptions import FailedActivity
from chaoslib.types import Configuration, Secrets
from logzero import logger

from .control import driver

__all__ = ["apply"]


def apply(
    configuration: Configuration = None,
    secrets: Secrets = None,
    **kwargs,
):
    driver.apply(**kwargs)



from typing import Any, Dict, List

from chaoslib.exceptions import InterruptExecution
from chaoslib.types import (
    Activity,
    Configuration,
    Experiment,
    Hypothesis,
    Journal,
    Run,
    Secrets,
    Settings,
)
from logzero import logger

from .driver import Terraform

VAR_NAME_PREFIX = "tf__"
CONFIG_PREFIX = "tf_conf__"
EXPORT_VAR_PREFIX = "tf_out__"


def configure_control(
    silent: bool = True,
    retain: bool = False,
    configuration: Configuration = None,
    secrets: Secrets = None,
    settings: Settings = None,
    experiment: Experiment = None,
):
    tf_vars = {}
    if configuration:
        for key, value in configuration.items():
            if key.startswith(VAR_NAME_PREFIX):
                tf_key = key.replace(VAR_NAME_PREFIX, "")
                tf_vars[tf_key] = value

    for key, _ in tf_vars.items():
        configuration.pop(f"{VAR_NAME_PREFIX}{key}")

    params = {
        "retain": bool(configuration.get(f"{CONFIG_PREFIX}retain", retain)),
        "silent": bool(configuration.get(f"{CONFIG_PREFIX}silent", silent)),
    }
    logger.info(
        "Terraform: retain stack after experiment completion: %s",
        str(params.get("retain")),
    )

    driver = Terraform()
    driver.configure(**params, args=tf_vars)
    driver.terraform_init()


def before_experiment_control(
    context: Experiment,
    configuration: Configuration = None,
    secrets: Secrets = None,
    **kwargs,
):
    """
    before-control of the experiment's execution

    Called by the Chaos Toolkit before the experiment's begin but after the
    configuration and secrets have been loaded.
    """
    driver = Terraform()
    logger.info("Terraform: creating required resources for experiment")
    driver.apply()
    for key, value in driver.output().items():
        logger.info("Terraform: reading configuration value for [%s]", key)
        configuration[f"{EXPORT_VAR_PREFIX}{key}"] = value.get("value")


def after_experiment_control(
    context: Experiment,
    state: Journal,
    configuration: Configuration = None,
    secrets: Secrets = None,
    **kwargs,
):
    driver = Terraform()
    if not driver.retain:
        logger.info("Terraform: removing experiment resources")
        driver.destroy()
    else:
        logger.info(
            "Terraform: stack resources will be retained after experiment completion."
        )
